#!/usr/local/bin/python

__author__     = "Aaron Straup Cope"
__url__        = "http://aaronland.info/python/w5"
__version__    = "1.0a8"
__cvsversion__ = "$Revision: 1.2 $"
__date__       = "$Date: 2004/11/09 15:15:52 $"
__copyright__  = "%s\n%s\n%s" % \
                 ("This application is licensed under the",
                  "Creative Commons Attribution-NonCommercial License",
                  "http://creativecommons.org/licenses/by-nc/1.0/")

__license__    = "This application is licensed under the Creative \
Commons Attribution-NonCommercial License. To view a copy of this \
license, visit http://creativecommons.org/licenses/by-nc/1.0/ or send \
a letter to Creative Commons, 559 Nathan Abbott Way, Stanford, \
California 94305, USA."

"""
This is a very simple CGI-based XML-RPC server for serving
images to the 'so-called w5 application'. It is also, currently,
the only server that will handle caching/checking of seen images
using Bloom Filters correctly. The Bloom Filter hooks do not work
across language implentations, specifically the Perl Bloom::Filter
module. This is considered an outstanding bug to be addressed.

---

The server requires two configuration variables :

 LOCAL_PHOTO_DIR : The path to the directory containing your images

 LOCAL_PHOTO_RE  : A regular expression used to match image files.
                   Example, match all JPEG, GIF and PNG images :
                    ".*\.(?:gif|jpg|png)$"

"""

LOCAL_PHOTO_DIR = ""
LOCAL_PHOTO_RE = ".*\.(?:jpg|gif|png)$";

import os
import re
import random
import xmlrpclib
import md5
import sys

from array import array
from struct import unpack
from base64 import encodestring, decodestring

from SimpleXMLRPCServer import CGIXMLRPCRequestHandler

# (This is the same a rImage/rBloom.py)
# To be properly imported using standard
# python hooks once I figure out python
# magic for installing libraries

class rBloom:
	LOAD_FACTOR = 8
	ITEMS       = 1024

        # we might eventually let users define
        # a ceiling but for now it is hard-coded
        
	def __init__(self,items=0):
		self.bits   = self.ITEMS * self.LOAD_FACTOR
		self.bitmap = array('B', '\0' * (self.bits >> 3))
	
	def add(self, pattern):

		m = md5.new()
		m.update(pattern)
		
		idxs = unpack('4I', m.digest())
		for i in idxs:
			i %= self.bits
			offset, mask = i >> 3, 1 << (i % 8)
			self.bitmap[offset] |= mask
	
	def match(self, pattern):
		m = md5.new()
		m.update(pattern)
		
		idxs = unpack('4I', m.digest())
		for i in idxs:
			i %= self.bits
			offset, mask = i >> 3, 1 << (i % 8)
			if self.bitmap[offset] & mask != mask:
				return False
		
		return True

	def fromstring (self, str):
            a = array('B')
            a.fromstring(str)
            self.bitmap = a
            
	def tostring (self):
	    return self.bitmap.tostring()

        def fromstring_encoded (self,str):
            return self.fromstring(decodestring(str))

        def tostring_encoded (self) :
            return encodestring(self.tostring())

# ibid. (for more easily subclassing)
# The main logic for retrieving images

class w5 :

    def list_images (self) :
        images = []
        
        is_jpeg = re.compile(LOCAL_PHOTO_RE)
        
        for root, dirs, files in os.walk(LOCAL_PHOTO_DIR):
            for name in files:
                
                if is_jpeg.match(name) :
                    images.append(os.path.join(root, name))
                    
        return images
                
    def retrieve_image (self,filter = None) :
        list   = self.list_images()
        result = None

        while result == None:
            file = random.choice(list)
            
            if filter == None:
                result = file
                break

            m = str(md5.new(file).hexdigest())
        
            if filter.match(m) != True:
                result = file
                break

            index = list.index(file)
            list.pop(index)

            if len(list) == 0:
                break

        return result

    def examine_image (self,img,data) :

        # check for JPEG-ness
        # check for JPEG comment
        # check for RDF-ness
        # set data with :
        # dc:title       (what)
        # dc:coverage    (where)
        # dc:depicts     (who)
        # dc:date        (when)
        # dc:description (why)
        
        pass

    def RandomImage (self,params={}):
        # sys.stderr.write("version %s" % params["release"])
        
        filter = None
        
        if params.has_key("filter") :
            filter = rBloom()        
            filter.fromstring_encoded(params["filter"])

        #
        
        img = self.retrieve_image(filter)

        if img == None :
            raise xmlrpclib.Fault(500,"unable to retrieve image")

        #
        
        res = {"who"   : "",
               "what"  : "",
               "where" : "",
               "when"  : "",
               "uid"   : "",
               "image" : ""}

        self.examine_image(img,res)
        
        res["uid"]   = md5.new(img).hexdigest()
        res["image"] = xmlrpclib.Binary(open(img).read())
        
        return res

    def _dispatch(self,method,params) :
        if re.compile("^(?:w5\.)?RandomImage$").match(method) :
            return self.RandomImage(params[0])
        else :
            raise 'Unknown method'

# main, e.g. the server itself

server = CGIXMLRPCRequestHandler()
server.register_instance(w5())
server.handle_request()
